#include "bezierPatch.h"

#include <cassert>
#include <windows.h>
#include <GL/gl.h>
#include <string>

////////////////////////////// TBezierPatch ////////////////////////////////////

TBezierPatch::TBezierPatch() :
	fDegreeU(0), fDegreeV(0), fValid(false)
{
	fSamples.resize(sqr(samplesPerSegment()));
}

TBezierPatch::~TBezierPatch() {
	// no code
}

TVector4d TBezierPatch::vertex(TIndex index) const {
	if (!index.isValid())
		throw std::string("Ivalid index");

	return fVertices.at(index.index());
}

void TBezierPatch::setVertex(TIndex index, TVector4d value) {
	if (!index.isValid())
		throw std::string("Ivalid index");

	fVertices.at(index.index()) = value;

	fValid = false;
}

unsigned int TBezierPatch::degree(TDirection direction) const {
	// NOTE - for Bezier triangles ignore direction parameter

	switch (direction) {
		case dU:
			return fDegreeU;
		case dV:
			return fDegreeV;
		default:
			assert(0);
						return 0;
	}
}

void TBezierPatch::setDegree(TDirection direction, unsigned int value) {
	// NOTE - for Bezier triangles ignore direction parameter

	switch (direction) {
		case dU:
			fDegreeU = value;
			reset();
		break;
		case dV:
			fDegreeV = value;
			reset();
		break;
		default:
			assert(0);
	}
}

double TBezierPatch::knot(TDirection dir, unsigned int index) {
	// NOTE - for patches which use knot vectors return specified knot

	return 0.0;
}

void TBezierPatch::setKnot(TDirection dir, unsigned int index, double value) {
	// NOTE - for patches which use knot vectors update specified knot

	// no code
}

unsigned int TBezierPatch::knotVectorSize(TDirection dir) {
	// NOTE - for patches which use knot vectors return size of the vector

	return 0;
}

TBezierPatch::TIndex TBezierPatch::getVertexNear(const TVector4d& linePoint,
	const TVector4d& lineVector, double maxRadius)
{
	if (fVertices.empty())
		return TIndex();

	double bestDistance = 10000; // big number
	double distance;
	unsigned int best;
	TVector4d point;
	double vectorSizeSq = sqr(lineVector.x()) + sqr(lineVector.y()) +
		sqr(lineVector.z());
	double t;

	// we iterate through all vertices
	for (unsigned int i = 0; i < fVertices.size(); i++) {
		// we compute parameter for the parametric line where it intersects
		// plane defined by currently processed control point and line
		// direction vector as normal vector
		t = (lineVector.x() * (fVertices[i].x() - linePoint.x()) +
			lineVector.y() * (fVertices[i].y() - linePoint.y()) +
			lineVector.z() * (fVertices[i].z() - linePoint.z())) / vectorSizeSq;

		// we obtain point on line belonging to previously computed parameter
		point = TVector4d(linePoint.x() + t * lineVector.x(),
			linePoint.y() + t * lineVector.y(), linePoint.z() +
			t * lineVector.z());

		// we compute the distance of control point from our point
		distance = sqr(point.x() - fVertices[i].x()) +
			sqr(point.y() - fVertices[i].y()) +
			sqr(point.z() - fVertices[i].z());

		// if the distance is smaller than current best, store it
		if (distance < bestDistance) {
			bestDistance = distance;
			best = i;
		}
	}

	// ignore distances larger than specified threshold
	if (bestDistance > sqr(maxRadius))
		return TIndex();

	return TIndex(true, best);
}

void TBezierPatch::samplesPerSegmentChanged(unsigned int oldValue,
	unsigned int newValue)
{
	// we resize the array for samples and unset the flag storing whether
	// samples are valid
	fSamples.resize(newValue * newValue);
	fValid = false;
}

void TBezierPatch::recompute() {
	if ((samplesPerSegment() < 2) || (degree(dU) * degree(dV) == 0))
		return;

	// we uniformly sample the domain of the patch
	unsigned int index = 0;
	for (unsigned int j = 0; j < samplesPerSegment(); j++) {
		for (unsigned int i = 0; i < samplesPerSegment(); i++) {
			// u & v will hold parameters of the current sample
			double u = i / (double)(samplesPerSegment() - 1);
			double v = j / (double)(samplesPerSegment() - 1);

			// TODO - use de Casteljau algorithm to get point on patch

			// NOTE - for this sample application we just compute linear
			// interpolation of nearest vertices
			unsigned int nearestU =
				static_cast<unsigned int>(degree(dU) * u);
			unsigned int nearestV =
				static_cast<unsigned int>(degree(dV) * v);

			double coefU = (std::min(nearestU + 1, degree(dU)) *
				(1.0 / degree(dU)) - u) * degree(dU);
			double coefV = (std::min(nearestV + 1, degree(dV)) *
				(1.0 / degree(dV)) - v) * degree(dV);

			TVector4d point =
				(fVertices.at(nearestV * (degree(dU) + 1) + nearestU) *
					coefU +
				fVertices.at(nearestV * (degree(dU) + 1) +
				std::min(nearestU + 1, degree(dU))) * (1.0 - coefU)) *
				coefV +
				(fVertices.at(std::min(nearestV + 1, degree(dV)) *
				(degree(dU) + 1) + nearestU) * coefU +
				fVertices.at(std::min(nearestV + 1, degree(dV)) *
				(degree(dU) + 1) + std::min(nearestU + 1, degree(dU))) *
				(1.0 - coefU)) *
				(1.0 - coefV);

			// TODO - end

			fSamples.at(index) = point;

			index++;
		}
	}

	fValid = true;
}

void TBezierPatch::paint(TIndex activeVertex, TParam highlight, TDrawMode drawMode)
{
		// if the current samples are not valid, compute new ones
		if (!fValid)
				recompute();


	glDisable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);

	// we draw the surface
	if (((drawMode & dmSolid) != dmNone) && (samplesPerSegment() >= 2)) {
		glEnable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

		glBegin(GL_TRIANGLES);
			for (unsigned int j = 0; j < samplesPerSegment() - 1; j++) {
				for (unsigned int i = 0; i < samplesPerSegment() - 1; i++) {
					unsigned int index = j * samplesPerSegment() + i;

					// we take 4 samples from the array
					TVector4d vertex1 = fSamples.at(index);
					TVector4d vertex2 = fSamples.at(index + 1);
					TVector4d vertex3 = fSamples.at(index +
						samplesPerSegment());
					TVector4d vertex4 = fSamples.at(index + 1 +
						samplesPerSegment());

					// We set the normal. You may compute the normal
					// analytically for each vertex
					TVector4d vector1 = vertex3 - vertex2;
					TVector4d vector2 = vertex1 - vertex2;

					glNormal3d(
						vector1.y() * vector2.z() - vector1.z() * vector2.y(),
						vector1.z() * vector2.x() - vector1.x() * vector2.z(),
						vector1.x() * vector2.y() - vector1.y() * vector2.x()
					);

					glVertex3d(vertex1.x(), vertex1.y(), vertex1.z());
					glVertex3d(vertex2.x(), vertex2.y(), vertex2.z());
					glVertex3d(vertex3.x(), vertex3.y(), vertex3.z());

					// We now set the normal for the second triangle.
					vector1 = -vector1;
					vector2 = vertex4 - vertex2;

					glNormal3d(
						vector1.y() * vector2.z() - vector1.z() * vector2.y(),
						vector1.z() * vector2.x() - vector1.x() * vector2.z(),
						vector1.x() * vector2.y() - vector1.y() * vector2.x()
					);

					glVertex3d(vertex3.x(), vertex3.y(), vertex3.z());
					glVertex3d(vertex2.x(), vertex2.y(), vertex2.z());
					glVertex3d(vertex4.x(), vertex4.y(), vertex4.z());
				}
			}
		glEnd();
		glDisable(GL_LIGHTING);
	}

	// We draw the wire-frame of the patch.
	if (((drawMode & dmWireFrame) != dmNone) && (samplesPerSegment() >= 2)) {
		glColor3f(0.5, 0.5, 0.5);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

		unsigned int index = 0;
		for (unsigned int j = 0; j < samplesPerSegment(); j++) {
				glBegin(GL_LINE_STRIP);
					for (unsigned int i = 0; i < samplesPerSegment(); i++) {
						glVertex3d(fSamples[index].x(), fSamples[index].y(),
							fSamples[index].z());

						index++;
					}
				glEnd();
		}

		index = 0;
		for (unsigned int i = 0; i < samplesPerSegment(); i++) {
			glBegin(GL_LINE_STRIP);
				for (unsigned int j = 0; j < samplesPerSegment(); j++) {
					glVertex3d(fSamples[index].x(), fSamples[index].y(),
						fSamples[index].z());

					index += samplesPerSegment();
				}
			glEnd();

			index = (index % samplesPerSegment()) + 1;
		}
	}

	if ((drawMode & dmGrid) != dmNone) {
		glColor3f(0.3, 0.3, 0.3);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

		// TODO Draw the wire-frame control patch.
	}

	// We draw the vertices
	if ((drawMode & dmVertices) != dmNone) {
		glDisable(GL_DEPTH_TEST);
		glColor3f(0.7, 0.7, 0.7);

		glBegin(GL_POINTS);
			for (unsigned int i = 0; i < fVertices.size(); i++)
				glVertex3d(fVertices.at(i).x(), fVertices.at(i).y(),
					fVertices.at(i).z());
		glEnd();

		glEnable(GL_DEPTH_TEST);
	}

	if ((drawMode & dmHighlight) != dmNone) {
		// TODO Draw point on the patch belonging to highlight parameter
		// and display also normal and tangents.
	}

	if ((drawMode & dmIsoLines) != dmNone) {
		// TODO Draw iso-lines for parameters from highlight.
	}
}

void TBezierPatch::reset() {
	fVertices.resize((degree(dU) + 1) * (degree(dV) + 1));

	unsigned int index = 0;
	for (unsigned int j = 0; j <= degree(dV); j++) {
		for (unsigned int i = 0; i <= degree(dU); i++) {
			fVertices.at(index) =
				TVector4d(i / (double)fDegreeU, j / (double)fDegreeV, 0.0, 1.0);

			index++;
		}
	}

	fValid = false;
}


