#ifndef PLOCHY_PATCH_H
#define PLOCHY_PATCH_H

#include "types.h"

class TPatch {
private:
	unsigned int fSamplesPerSegment;
protected:
	/// Called when number of samples per segment changes.
	virtual void samplesPerSegmentChanged(unsigned int oldValue,
		unsigned int newValue);
public:
	/// Type representing index of a control vertex
	class TIndex {
	private:
		bool fValid;
		unsigned int fIndex;
	public:
		TIndex(bool valid = false, unsigned int index = 0);
		bool isValid() const;
		unsigned int index() const;
		bool operator==(TIndex &copy) const;
		bool operator!=(TIndex &copy) const;
	};

	/// Parameter triplet representing point in domain of the patch.
	/// In all cases sum of elements equals '1'.
	class TParam {
	private:
		double fU, fV, fW;
	public:
		TParam(double u = 0.0, double v = 0.0);
		double u() const;
		double v() const;
		double w() const;
		void setU(double value);
		void setV(double value);
		void setW(double value);
	};

	/// Enum specifying which features should be rendered. Values may be or-ed
	/// 	together to select multiple features.
	/// dmNone - nothing will be rendered
	/// dmWireFrame - wireframe model of the patch will be rendered
	/// dmSurface - shaded surface of the patch will be rendered
	/// dmVertices - control points will be displayed
	/// dmGrid - wireframe model of the patch with 0 as z-coordinate will be
	/// 	rendered
	/// dmHighlight - specified point on patch will be highlighted
	/// dmIsoLines - iso-lines for the specified point will be rendered
	/// dmAll - all features will be rendered
	enum TDrawMode {dmNone = 0, dmWireFrame = 1, dmSolid = 2, dmVertices = 4,
		dmGrid = 8, dmHighlight = 16, dmIsoLines = 32,
		dmAll = dmWireFrame | dmSolid | dmVertices | dmGrid | dmHighlight |
		dmIsoLines
	};

	enum TDirection {dU, dV};

	TPatch();
	virtual ~TPatch();
	/// Number of samples per segment of the patch. This value is used when
	/// rendering the patch.
	unsigned int samplesPerSegment() const;
	/// Sets number of samples per segment of the patch for rendering.
	void setSamplesPerSegment(unsigned int value);
	/// Returns coordinates of a control point with specified index.
	virtual TVector4d vertex(TIndex index) const = 0;
	/// Sets coordinates of a control point with specified index.
	virtual void setVertex(TIndex index, TVector4d value) = 0;
	/// Returns degree of the patch.
	virtual unsigned int degree(TDirection direction) const = 0;
	/// Sets degree of the patch.
	virtual void setDegree(TDirection direction, unsigned int value) = 0;
	/// Returns knot from knot vector.
	virtual double knot(TDirection dir, unsigned int index) = 0;
	/// Sets knot in knot vector.
	virtual void setKnot(TDirection dir, unsigned int index, double value) = 0;
	/// Returns knot vector size.
	virtual unsigned int knotVectorSize(TDirection dir) = 0;
	/// Returns index of a control point nearest to the specified line.
	virtual TIndex getVertexNear(const TVector4d& linePoint,
		const TVector4d& lineVector, double maxRadius = 0.05) = 0;
	/// Paints the patch according to the specified styles.
	virtual void paint(TIndex activeVertex = TIndex(),
		TParam highlight = TParam(), TDrawMode drawMode = dmAll) = 0;
};

#endif

